#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for casa_envoy.foundation
"""
from pathlib import Path
from unittest.mock import patch

import pytest
from casa_envoy.foundation import RestoreFoundation, GeneralFoundation, CubeFoundation, set_permissions
from casa_envoy.schema import AbstractTextFile

parameters = {
    "useCasa": False,
    "homeForReprocessing": "/home/casa/packages/pipeline/current",
    "rootDirectory": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool",
    "processingDirectory": "tmpo1ca1pp_",
    "parent_path": Path("/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpo1ca1pp_"),
    "metadata": "test/input_files/restore.json",
    "ppr": "test/input_files/PPR.xml",
    "product_type": "restore",
}

test_general_metadata = AbstractTextFile(
    filename="test/input_files/test-general-foundation.json",
    content=Path("test/input_files/test-general-foundation.json").read_text(),
)

test_restore_metadata = AbstractTextFile(
    filename="test/input_files/restore.json",
    content=Path("test/input_files/restore.json").read_text(),
)

general = GeneralFoundation(parameters=parameters, metadata=test_general_metadata)
foundation = RestoreFoundation(parameters=parameters, metadata=test_restore_metadata)
cube = CubeFoundation(parameters=parameters)


@patch("os.path.join")
@patch("os.chmod")
@patch("os.walk")
def test_set_permissions(mock_walk, mock_chmod, mock_join):
    """
    Verifies reusable set permissions function

    :param mock_walk:
    :param mock_chmod:
    :param mock_join:
    :return:
    """
    path = Path("I/am/a/path")
    set_permissions(path)
    assert mock_walk.call_count == 1


class TestGeneralFoundation:
    """
    Test methods for laying General data foundations
    """

    @patch("shutil.copy")
    def test_general_data_foundation(self, mock_copy):
        general.data_foundation()
        assert mock_copy.call_count == 2


class TestRestoreFoundation:
    """
    Test methods for laying Restore data foundations
    """

    @patch("os.chdir")
    def test_restore_data_foundation(self, mock_chdir):
        with patch("casa_envoy.foundation.RestoreFoundation.extract_cal") as extract:
            with patch("casa_envoy.foundation.set_permissions") as permissions:
                foundation.data_foundation()
                assert mock_chdir.call_count == 1

    @pytest.mark.skip("skipping because this test can't seem to tell the difference between a list and a string...")
    @patch("pathlib.Path.exists", return_value=True)
    @patch("tarfile.open")
    def test_extract_cal(self, mock_tar, mock_path):
        foundation.extract_cal()
        assert mock_tar.call_count == 1


class TestCubeFoundation:
    """
    Test methods for laying VLASS Cube data foundations
    """

    @patch("os.chdir")
    def test_cube_data_foundation(self, mock_chdir):
        with patch("casa_envoy.foundation.CubeFoundation.extract_reimaging_resources") as extract:
            with patch("casa_envoy.foundation.set_permissions") as permissions:
                cube.data_foundation()
                assert mock_chdir.call_count == 1
                assert extract.call_count == 1
                assert permissions.call_count == 1

    @patch("tarfile.open")
    @patch("pathlib.Path.exists", return_value=True)
    def test_extract_reimaging_resources(self, mock_exists, mock_tar_open):
        cube.extract_reimaging_resources()
        assert mock_exists.call_count == 1
        assert mock_tar_open.call_count == 1

    @patch("pathlib.Path.exists", return_value=True)
    @patch("shutil.copy")
    def test_move_parameters_list(self, mock_copy, mock_exists):
        cube.move_parameters_list()
        assert mock_exists.call_count == 1
        assert mock_copy.call_count == 1
