/*
 * Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */
import { Component, OnDestroy, OnInit } from "@angular/core";

import { CapabilityRequest } from "../../model/capability-request";
import { ActivatedRoute, Router } from "@angular/router";
import { HttpClient,HttpParams } from '@angular/common/http';
import { Title } from "@angular/platform-browser";
import { DataRetrieverService } from "../../services/data-retriever.service";
import { CapabilityExecution } from "../../model/capability-execution";
import { CapabilityRequestService } from "../../services/capability-request.service";
import { CapabilityVersion } from "../../model/capability-version";
import { PollingDataUpdaterService } from "../../services/polling-data-updater.service";
import { UrlService } from "../../services/url.service";
import { AlertService } from "../../../shared/alert/alert.service";
import { JsonObject } from "@angular/compiler-cli/ngcc/src/packages/entry_point";
import { Subject, Observable } from "rxjs";
import { takeUntil, repeatWhen } from "rxjs/operators";
import { Capability } from "../../model/capability";

@Component({
  selector: "app-capability-request",
  templateUrl: "./capability-request.component.html",
  styleUrls: ["./capability-request.component.scss"],
})
export class CapabilityRequestComponent implements OnInit, OnDestroy {
  public baseTitle = 'Capability Request #';
  public capabilityRequest: CapabilityRequest;
  public capability: Capability;
  public capabilityExecution: CapabilityExecution;
  public currentVersion: CapabilityVersion;
  public latestVersion: CapabilityVersion;
  public previousUrlString: string
  activeRequestsPageDefaultUrl: string = "/workspaces/active-requests"
  previousUrl: Observable<string> = this.urlService.previousUrl$;

  // public qaNotes: string;

  private ngUnsubscribe = new Subject(); // to signal when all subscribers should unsubscribe (triggered in tear down)

  // Observer for capability request objects
  private capabilityRequestObserver = {
    next: (request) => {
      this.capabilityRequest = request;
      if (!this.latestVersion) {
        this.latestVersion = this.capabilityRequest.versions[
          this.capabilityRequest.versions.length - 1
        ];
      }
      if (!this.currentVersion) {
        this.currentVersion = this.latestVersion;
      }
      if (this.currentVersion.current_execution) {
        this.capabilityExecution = this.currentVersion.current_execution;
      } else {
        // No current execution; use placeholder instead to avoid errors
        this.capabilityExecution = {
          capability_request_id: this.capabilityRequest.id,
          state_name: null,
          current_workflow_request_id: null,
          id: null,
          queue_state: null,
          version_number: null,
          delivery_url: null,
          created_at: null,
          updated_at: null,
          type: "CapabilityExecution",
        };
      }

      // Set page title
      this.titleService.setTitle(this.baseTitle + this.capabilityRequest.id)

      this.dataRetriever
        .getCapability(this.capabilityRequest.capability_name)
        .pipe(takeUntil(this.ngUnsubscribe))
        .subscribe(this.capabilityObserver);
    },
    error: (error) => console.error("Error when retrieving capability request:" + error),
  };
  // Observer for capability objects
  private capabilityObserver = {
    next: (capability) => {
      this.capability = capability;
    },
    error: (error) => console.error("Error when retrieving capability:" + error),
  };

  // Observer for capability version objects
  private capabilityVersionObserver = {
    next: (capabilityVersion) => {
      if (capabilityVersion) {
        this.latestVersion = capabilityVersion;
        // Set selected version to be the request's latest version
        // If no version is selected or if the selected version is the latest version
        if (
          !this.currentVersion ||
          this.currentVersion.version_number === this.latestVersion.version_number
        ) {
          this.setVersion(this.latestVersion);
        }
      } else {
        console.error("Current version returned undefined.");
      }
    },
    error: (error) => console.error("Error when retrieving current version:" + error),
  };

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private dataRetriever: DataRetrieverService,
    private capabilityRequestService: CapabilityRequestService,
    private pollingDataUpdaterService: PollingDataUpdaterService,
    private alertService: AlertService,
    private titleService: Title,
    private urlService: UrlService,
  ) {
    const requestID = parseInt(this.route.snapshot.paramMap.get("id"));
    this.pollingDataUpdaterService
      .getDataPoller$(this.capabilityRequestService.getCapabilityRequestURL(requestID))
      .pipe(
        takeUntil(this.ngUnsubscribe),
        repeatWhen(() => this.ngUnsubscribe)
      )
      .subscribe(this.capabilityRequestObserver);
    this.pollingDataUpdaterService
      .getDataPoller$(this.capabilityRequestService.getLatestVersionURL(requestID))
      .pipe(
        takeUntil(this.ngUnsubscribe),
        repeatWhen(() => this.ngUnsubscribe)
      )
      .subscribe(this.capabilityVersionObserver);
  }

  alertMessage(eventDetails: JsonObject): void {
    let message = "";
    let title = "";

    switch (eventDetails.type) {
      case "carta": {
        title = "Launching CARTA";
        message =
          "An email has been sent to <strong>" +
          eventDetails.email +
          "</strong> with a link to your CARTA instance.";
        this.alertService.info(message, title);
        break;
      }
      case "error": {
        message = "There has been an error with your request";
        title = "Error";
        this.alertService.error(message, title);
        break;
      }
    }
  }

  setVersion(version: CapabilityVersion): void {
    this.currentVersion = version;
  }

  catchOpenEditorEventFromChild(isEditorOpen: boolean): void {
    let shouldPausePoller = !isEditorOpen
    this.ngUnsubscribe.next(shouldPausePoller);
  }

  /**
   * Determine list of state transitions for a capability's state machine
   *
   * @param version
   */
  public determineQAable(version: CapabilityVersion){
    if (!version.current_execution)
      return false

    return version.current_execution.state_name === 'Awaiting QA' || version.current_execution.state_name === 'AoD Review';

  }

  public goToActiveRequestsPage() {
    this.ngOnDestroy()
    this.previousUrlString = (this.previousUrlString) ? this.previousUrlString : `${this.activeRequestsPageDefaultUrl}?capability=${this.capabilityRequest.capability_name}`
    this.router.navigateByUrl(this.previousUrlString)
  }

  ngOnInit(): void {
    this.urlService.previousUrl$.subscribe((previousUrl: string) => {
      this.previousUrlString = previousUrl
    });
  }

  ngOnDestroy(): void {
    // Prevent memory leakage by unsubscribing all observers
    this.ngUnsubscribe.next();
    this.ngUnsubscribe.complete();
  }
}
