#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=C0114, C0301, E0401, R0903, W1203
import copy
import logging
from typing import Dict, List

import transaction
from messaging.messenger import MessageSender
from messaging.router import Router, on_message

from workspaces.capability.enums import CapabilityVersionState
from workspaces.capability.helpers import Parameter
from workspaces.capability.message_architect import CapabilityMessageArchitect
from workspaces.capability.schema import CapabilityExecution, CapabilityVersion
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.execution_manager import ExecutionManager
from workspaces.capability.services.interfaces import CapabilityServiceIF
from workspaces.capability.services.state_machine_info import StateMachineInfo
from workspaces.notification.services.interfaces import NotificationServiceIF
from workspaces.workflow.services.interfaces import WorkflowServiceIF

logger = logging.getLogger(__name__)


class CapabilityService(CapabilityServiceIF):
    """
    The capability service: clients access this to request capability runs
    """

    def __init__(
        self,
        capability_info: CapabilityInfo,
        state_machine_info: StateMachineInfo,
        workflow_service: WorkflowServiceIF,
        notification_service: NotificationServiceIF,
    ):
        self.messenger = MessageSender("capability")
        self.router = Router("capability")
        self.router.register(self)
        self.execution_manager = ExecutionManager(
            capability_info, workflow_service, self.router, self.messenger, notification_service
        )
        self.capability_info = capability_info
        self.state_machine_info = state_machine_info
        self.state_machine_info.init_state_machines()

    def create_new_version(self, capability_request_id: int, parameters: List[Parameter] = None) -> CapabilityVersion:
        logger.info(f"Creating new version of Capability Request #{capability_request_id}")
        return self.capability_info.create_capability_version(capability_request_id, parameters)

    @on_message(type="execution-complete")
    def complete_request(self, **message: Dict):
        logger.info(f"RECEIVED EXECUTION-COMPLETE: {message}")

        execution = message["subject"]
        capability_request = self.capability_info.lookup_capability_request(execution["capability_request_id"])
        version = self.capability_info.lookup_version(capability_request.id, execution["version_number"])

        # Set version state to Complete if the execution is indeed complete; else, leave it alone
        version.state = CapabilityVersionState.Complete.name if execution["state_name"] == "Complete" else version.state
        self.capability_info.save_entity(version)

        # Update request state
        capability_request.determine_state()
        self.capability_info.save_entity(capability_request)

        capability_complete_msg = CapabilityMessageArchitect(version=version).compose_message("capability_complete")
        self.messenger.send_message(**capability_complete_msg)

    @on_message(type="workflow-aborted")
    @on_message(type="execution-failed")
    def fail_request(self, **message: Dict):
        logger.info(f"RECEIVED EXECUTION-FAILED: {message}")

        if message["type"] == "workflow-aborted":
            workflow_req = message["subject"]
            execution = self.capability_info.lookup_execution_by_workflow_request_id(
                workflow_req["workflow_request_id"]
            )
            if execution:
                cap_request_id = execution.capability_request_id
                cap_version_number = execution.version_number
            else:
                logger.info(f"No Execution found for {workflow_req['workflow_request_id']}")
        else:
            execution = message["subject"]
            cap_request_id = execution["capability_request_id"]
            cap_version_number = execution["version_number"]

        capability_request = self.capability_info.lookup_capability_request(cap_request_id)
        version = self.capability_info.lookup_version(capability_request.id, cap_version_number)

        # Set version state to Failed
        version.state = CapabilityVersionState.Failed.name
        self.capability_info.save_entity(version)

        # Update request state
        capability_request.determine_state()
        self.capability_info.save_entity(capability_request)
        capability_complete_msg = CapabilityMessageArchitect(version=version).compose_message("capability_failed")
        self.messenger.send_message(**capability_complete_msg)

    @on_message(type="capability-cancelled")
    def cancel_request(self, **message: Dict):
        logger.info(f"RECEIVED EXECUTION-CANCELLED: {message}")

        execution = message["subject"]
        capability_request = self.capability_info.lookup_capability_request(execution["capability_request_id"])
        version = self.capability_info.lookup_version(capability_request.id, execution["version_number"])

        # Set version state to Cancelled
        version.state = CapabilityVersionState.Cancelled.name
        self.capability_info.save_entity(version)

        # Update request state
        capability_request.determine_state()
        capability_request.sealed = True
        self.capability_info.save_entity(capability_request)

    @on_message(type="execution-error")
    def error_request(self, **message: Dict):
        logger.info(f"RECEIVED EXECUTION-ERROR: {message}")

        execution = message["subject"]
        capability_request = self.capability_info.lookup_capability_request(execution["capability_request_id"])
        version = self.capability_info.lookup_version(capability_request.id, execution["version_number"])

        # Set version state to Failed
        version.state = CapabilityVersionState.Error.name
        self.capability_info.save_entity(version)

        # Update request state
        capability_request.determine_state()
        self.capability_info.save_entity(capability_request)

        capability_complete_msg = CapabilityMessageArchitect(version=version).compose_message("capability_failed")
        self.messenger.send_message(**capability_complete_msg)

    @on_message(service="workflow", type="update-wf-metadata")
    def update_wf_metadata(self, **message: Dict):
        """
        Populate the workflow_metadata column for the workflow execution with
        the provided payload information for retrieval in the front-end

        :param message: an update-wf-metadata type message with a workflow request
        """
        json_payload = message["payload"]
        subject = message["subject"]

        execution = self.capability_info.lookup_execution_by_workflow_request_id(subject["workflow_request_id"])

        if execution is None:
            return

        # Update the database
        execution.version.workflow_metadata = copy.deepcopy(execution.version.workflow_metadata)

        if execution.version.workflow_metadata is None:
            execution.version.workflow_metadata = {}

        for key in json_payload:
            execution.version.workflow_metadata[key] = json_payload.get(key)

    @on_message(service="workflow", type="delivery")
    def on_delivery(self, **message: Dict):
        """
        Catch a delivery notification and update the locations in the
        referenced execution based on it

        :param message: a delivery-type notification with a workflow request
        subject and a delivery field
        """
        logger.info("RECEIVED DELIVERY NOTIFICATION: %s", message)

        # retrieve the delivery structure in the message
        delivery = message["delivery"]

        # we caught a delivery event, find the associated capability execution
        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            message["subject"]["workflow_request_id"]
        )

        # update some fields on the execution, if possible
        # Extra / at the end added to make nginx happy  -nhertz
        execution.delivery_url = delivery["url"] + "/" if "url" in delivery else None
        execution.delivery_path = delivery["delivered_to"] if "delivered_to" in delivery else None

    @on_message(type="capability-submitted")
    def execute_submitted_capability(self, **message: Dict):
        logger.info(f"Executing capability: {message}")
        subject = message["subject"]
        execution = self.capability_info.lookup_execution(subject["id"])
        self.execution_manager.execute(execution)

    @on_message(service="workflow", type="ingestion-complete")
    def on_ingestion_complete(self, **message: Dict[str, str]):
        """
        Catch an ingestion-complete message sent from ingest_envoy and update the ingested flag
        of the corresponding capability request

        :param message: Ingestion-complete message
        """
        logger.info(f"RECEIVED INGESTION COMPLETE MESSAGE: {message}")

        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            int(message["subject"]["execution_wf_id"])
        )

        execution.version.state = CapabilityVersionState.Complete.name
        self.capability_info.save_entity(execution.version)

        request = self.capability_info.lookup_capability_request(execution.capability_request.id)
        request.ingested = True
        request.sealed = True
        self.capability_info.save_entity(request)

    @on_message(type="carta-ready")
    def on_carta_ready(self, **message: Dict[str, str]):
        """
        Catch the RH-flavored event and save it to the capability request version metadata

        :param message: Ingestion-complete message
        :return:
        """
        logger.info(f"RECEIVED CARTA READY MESSAGE: {message}")

        subject = message["subject"]
        parent_wf_request_id = int(subject["argument"]["parent_wf_request_id"])

        execution = self.capability_info.lookup_execution_by_workflow_request_id(parent_wf_request_id)
        request_version = execution.version
        workflow_info = request_version.workflow_metadata
        if workflow_info:
            request_version.workflow_metadata = {**workflow_info, "carta_url": message["carta_url"]}
        else:
            request_version.workflow_metadata = {"carta_url": message["carta_url"]}

        self.capability_info.save_entity(request_version)

    @on_message(service="capability", type="restart-queue")
    def restart_queue(self, **message: Dict):
        # Restart queue manager
        self.execution_manager.queue_manager.startup()

    @on_message(type="capability-complete")
    @on_message(type="capability-failed")
    @on_message(type="qa_ready")
    @on_message(service="workflow", type="ingestion-failed")
    def log_message(self, **message: Dict):
        """
        Log a message that formerly produced a notification. These include `capability-complete`, `capability-failed`,
        and `qa_ready` from the capability service, and `ingestion-failed` via the workflow service. Nothing really
        needs to happen here, but it may be useful for debugging reasons.

        If it isn't useful, we should just remove this entire block.

        :param message:  the message we received
        """
        logger.debug(f"Received {message['type']} (no action taken): {message}")


class CapabilityLauncher:
    """
    The CapabilityLauncher provides a way for REST clients to access the server-singleton
    CapabilityService for basically the one thing they need it for, to launch capabilities.

    It's important that we share the message router here with the global one, as otherwise
    messages get round-robined to requests-based CapabilityLaunchers rather than the server-
    singleton CapabilityService which actually knows how to process these events.
    """

    def __init__(self, capability_info: CapabilityInfo, messenger: MessageSender):
        # since this is used by the REST API to implement its methods,
        # it's fairly important that it not register itself for on_message replies.
        logger.info("Instantiating capability launcher")
        self.messenger = messenger
        self.capability_info = capability_info

    def launch_capability(self, version: CapabilityVersion) -> CapabilityExecution:
        """
        Use the interface of the execution manager to run a capability version

        :param version: Capability version
        :return: Execution of the version
        """
        logger.info("Launching %s request #%s v%s", version.capability_name, version.request.id, version.version_number)
        # first we must create the execution
        execution = self.capability_info.create_execution(version)
        # Mark version as running
        execution.version.state = CapabilityVersionState.Running.name
        execution.version.sealed = True

        if execution.capability.single_version_only and execution.version.version_number == 1:
            request = self.capability_info.lookup_capability_request(execution.capability_request_id)
            request.sealed = True
            self.capability_info.save_entity(request)

        # ensure it's saved
        transaction.commit()

        # now pass it back to the CapabilityService to do the next thing
        capability_submitted_msg = CapabilityMessageArchitect(execution=execution).compose_message(
            "capability_submitted"
        )
        self.messenger.send_message(**capability_submitted_msg)

        return execution


class CapabilityQueueStarter:
    """
    Start the capability queue
    """

    def __init__(self, messenger: MessageSender):
        # since this is used by the REST API to implement its methods,
        # it's fairly important that it not register itself for on_message replies.
        logger.info("Instantiating capability queue starter")
        self.messenger = messenger

    def restart_capability_queue(self):
        """
        Tells the capability queue to restart
        """
        # now pass it back to the CapabilityService to do the next thing

        msg = CapabilityMessageArchitect(request=None).compose_message("restart_queue")
        self.messenger.send_message(**msg)


class CapabilityCanceller:
    """
    Start the capability queue
    """

    def __init__(self, messenger: MessageSender):
        # since this is used by the REST API to implement its methods,
        # it's fairly important that it not register itself for on_message replies.
        logger.info("Instantiating capability canceller")
        self.messenger = messenger

    def cancel_capability_execution(self, execution: CapabilityExecution):
        """
        Sends cancel message to cancel capability
        """

        cancel_msg = CapabilityMessageArchitect(execution=execution).compose_message("cancel")

        self.messenger.send_message(**cancel_msg)
